
/* Chrysalide - Outil d'analyse de fichiers binaires
 * base64.c - transormation en encodage Base64
 *
 * Copyright (C) 2023 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "base64.h"


#include <assert.h>
#include <malloc.h>
#include <stdint.h>
#include <string.h>


#include <analysis/scan/patterns/modifier-int.h>


#include "../base64.h"



/* ----------------------- RECHERCHE D'UN MOTIF DE TEXTE BRUT ----------------------- */


/* Initialise la classe des transmissions en encodage Base64. */
static void g_scan_base64_modifier_class_init(GScanBase64ModifierClass *klass);

/* Initialise une instance de transmission en encodage Base64. */
static void g_scan_base64_modifier_init(GScanBase64Modifier *);

/* Supprime toutes les références externes. */
static void g_scan_base64_modifier_dispose(GScanBase64Modifier *);

/* Procède à la libération totale de la mémoire. */
static void g_scan_base64_modifier_finalize(GScanBase64Modifier *);



/* --------------------- IMPLEMENTATION DES FONCTIONS DE CLASSE --------------------- */


/* Fournit le nom d'appel d'un modificateur pour motif. */
static char *g_scan_base64_modifier_get_name(const GScanBase64Modifier *);

/* Finalise l'encoddage en Base64 d'un motif transformé. */
static void strip_base64_modifier_output(const sized_binary_t *, size_t, sized_binary_t *);

/* Transforme une séquence d'octets pour motif de recherche. */
static bool g_scan_base64_modifier_transform(const GScanBase64Modifier *, const sized_binary_t *, size_t, sized_binary_t **, size_t *);

/* Détermine si un argument est bien toléré par un modificateur. */
static bool g_scan_base64_modifier_can_handle_arg(const GScanBase64Modifier *, const modifier_arg_t *);

/* Transforme une séquence d'octets pour motif de recherche. */
static bool g_scan_base64_modifier_transform_with_arg(const GScanBase64Modifier *, const sized_binary_t *, size_t, const modifier_arg_t *, sized_binary_t **, size_t *);



/* ---------------------------------------------------------------------------------- */
/*                         RECHERCHE D'UN MOTIF DE TEXTE BRUT                         */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini pour une transormation en encodage Base64. */
G_DEFINE_TYPE(GScanBase64Modifier, g_scan_base64_modifier, G_TYPE_SCAN_TOKEN_MODIFIER);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des transmissions en encodage Base64.   *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_base64_modifier_class_init(GScanBase64ModifierClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GScanTokenModifierClass *modifier;      /* Version de classe parente   */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_scan_base64_modifier_dispose;
    object->finalize = (GObjectFinalizeFunc)g_scan_base64_modifier_finalize;

    modifier = G_SCAN_TOKEN_MODIFIER_CLASS(klass);

    modifier->get_name = (get_scan_modifier_name_fc)g_scan_base64_modifier_get_name;

    modifier->transform = (transform_scan_token_fc)g_scan_base64_modifier_transform;
    modifier->can_handle = (can_token_modifier_handle_arg)g_scan_base64_modifier_can_handle_arg;
    modifier->transform_with = (transform_scan_token_with_fc)g_scan_base64_modifier_transform_with_arg;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : modifier = instance à initialiser.                           *
*                                                                             *
*  Description : Initialise une instance de transmission en encodage Base64.  *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_base64_modifier_init(GScanBase64Modifier *modifier)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : modifier = instance d'objet GLib à traiter.                  *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_base64_modifier_dispose(GScanBase64Modifier *modifier)
{
    G_OBJECT_CLASS(g_scan_base64_modifier_parent_class)->dispose(G_OBJECT(modifier));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : modifier = instance d'objet GLib à traiter.                  *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_base64_modifier_finalize(GScanBase64Modifier *modifier)
{
    G_OBJECT_CLASS(g_scan_base64_modifier_parent_class)->finalize(G_OBJECT(modifier));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Construit un modificateur livrant des encodages en Base64.   *
*                                                                             *
*  Retour      : Mécanisme mis en place.                                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GScanTokenModifier *g_scan_base64_modifier_new(void)
{
    GScanTokenModifier *result;                    /* Structure à retourner       */

    result = g_object_new(G_TYPE_SCAN_BASE64_MODIFIER, NULL);

    return result;

}



/* ---------------------------------------------------------------------------------- */
/*                       IMPLEMENTATION DES FONCTIONS DE CLASSE                       */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : modifier = modificateur à consulter.                         *
*                                                                             *
*  Description : Fournit le nom d'appel d'un modificateur pour motif.         *
*                                                                             *
*  Retour      : Désignation humaine.                                         *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static char *g_scan_base64_modifier_get_name(const GScanBase64Modifier *modifier)
{
    char *result;                           /* Désignation à retourner     */

    result = strdup("base64");

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : input  = encodage en Base64 obtenu.                          *
*                skip   = nombre de caractères initiaux à sauter.             *
*                output = encodage en Base64 final à conserver.               *
*                                                                             *
*  Description : Finalise l'encoddage en Base64 d'un motif transformé.        *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void strip_base64_modifier_output(const sized_binary_t *input, size_t skip, sized_binary_t *output)
{
    size_t final_len;                       /* Taille de représentation    */

    final_len = input->len;

    if (final_len > 0 && input->bin_data[final_len - 1] == '=')
    {
        while (final_len > 0 && input->bin_data[final_len - 1] == '=')
            final_len--;

        final_len--;

    }

    if (skip > 0)
        skip++;

    assert(final_len >= skip);

    final_len -= skip;

    output->len = final_len;
    output->bin_data = malloc(final_len * sizeof(bin_t));

    memcpy(output->bin_data, input->static_bin_data + skip, final_len);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : modifier = modificateur à solliciter.                        *
*                src      = séquences d'octets à traiter.                     *
*                scount   = quantité de ces séquences.                        *
*                dest     = nouvelle(s) séquence(s) d'octets obtenue(s) [OUT] *
*                dcount   = quantité de ces séquences.                        *
*                                                                             *
*  Description : Transforme une séquence d'octets pour motif de recherche.    *
*                                                                             *
*  Retour      : Bilan de l'opération : succès ou échec.                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_scan_base64_modifier_transform(const GScanBase64Modifier *modifier, const sized_binary_t *src, size_t scount, sized_binary_t **dest, size_t *dcount)
{
    bool result;                            /* Bilan d'opération à renvoyer*/
    sized_binary_t *binary;                 /* Raccourci vers le stockage  */
    size_t i;                               /* Boucle de parcours #1       */
    const sized_binary_t *_src;             /* Source courante             */
    sized_binary_t tmp_in;                  /* Manipulation des variations */
    sized_binary_t tmp_out;                 /* Manipulation des variations */

    result = true;

    *dcount = scount * 3;
    *dest = calloc(*dcount, sizeof(sized_binary_t));

    binary = &(*dest)[0];

    for (i = 0; i < scount && result; i++)
    {
        _src = src + i;

        tmp_in.data = malloc((_src->len + 2) * sizeof(bin_t));

        /**
         * Explications sur la définition des modifications sur les
         * encodages produits : https://www.leeholmes.com/searching-for-content-in-base-64-strings/
         */

        /* Encodage premier */

        result = base64_encode(_src, &tmp_out);
        if (!result) goto exit;

        strip_base64_modifier_output(&tmp_out, 0, binary++);

        exit_szstr(&tmp_out);

        /* Variante 1 */

        tmp_in.data[0] = ' ';
        memcpy(&tmp_in.data[1], _src->static_bin_data, _src->len);

        tmp_in.len = _src->len + 1;

        result = base64_encode(&tmp_in, &tmp_out);
        if (!result) goto exit;

        strip_base64_modifier_output(&tmp_out, 1, binary++);

        exit_szstr(&tmp_out);

        /* Variante 2 */

        tmp_in.data[0] = ' ';
        tmp_in.data[2] = ' ';
        memcpy(&tmp_in.data[2], _src->static_bin_data, _src->len);

        tmp_in.len = _src->len + 2;

        result = base64_encode(&tmp_in, &tmp_out);
        if (!result) goto exit;

        strip_base64_modifier_output(&tmp_out, 2, binary++);

        exit_szstr(&tmp_out);

 exit:

        exit_szstr(&tmp_in);

    }

    if (!result)
    {
        for (i = 0; i < *dcount; i++)
            exit_szstr(&(*dest)[i]);

        free(*dest);

        *dcount = 0;
        *dest = NULL;

    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : modifier = modificateur à solliciter.                        *
*                arg      = argument de personnalisation.                     *
*                                                                             *
*  Description : Détermine si un argument est bien toléré par un modificateur.*
*                                                                             *
*  Retour      : Bilan de la consultation : support ou non.                   *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_scan_base64_modifier_can_handle_arg(const GScanBase64Modifier *modifier, const modifier_arg_t *arg)
{
    bool result;                            /* Bilan d'opération à renvoyer*/

    switch (arg->type)
    {
        case MAT_STRING:
            result = (arg->value.string.len == 64);
            break;

        default:
            result = false;

    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : modifier = modificateur à solliciter.                        *
*                src      = séquences d'octets à traiter.                     *
*                scount   = quantité de ces séquences.                        *
*                arg      = argument de personnalisation.                     *
*                dest     = nouvelle(s) séquence(s) d'octets obtenue(s) [OUT] *
*                dcount   = quantité de ces séquences.                        *
*                                                                             *
*  Description : Transforme une séquence d'octets pour motif de recherche.    *
*                                                                             *
*  Retour      : Bilan de l'opération : succès ou échec.                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_scan_base64_modifier_transform_with_arg(const GScanBase64Modifier *modifier, const sized_binary_t *src, size_t scount, const modifier_arg_t *arg, sized_binary_t **dest, size_t *dcount)
{
    bool result;                            /* Bilan d'opération à renvoyer*/
    sized_binary_t *binary;                 /* Raccourci vers le stockage  */
    size_t i;                               /* Boucle de parcours #1       */
    const sized_binary_t *_src;             /* Source courante             */
    sized_binary_t tmp_in;                  /* Manipulation des variations */
    sized_binary_t tmp_out;                 /* Manipulation des variations */

    result = (arg->type == MAT_STRING);
    if (!result) goto bad_arg;

    *dcount = scount * 3;
    *dest = calloc(*dcount, sizeof(sized_binary_t));

    binary = &(*dest)[0];

    for (i = 0; i < scount && result; i++)
    {
        _src = src + i;

        tmp_in.data = malloc((_src->len + 2) * sizeof(bin_t));

        /**
         * Explications sur la définition des modifications sur les
         * encodages produits : https://www.leeholmes.com/searching-for-content-in-base-64-strings/
         */

        /* Encodage premier */

        result = _base64_encode(_src, &tmp_out, &arg->value.string);
        if (!result) goto exit;

        strip_base64_modifier_output(&tmp_out, 0, binary++);

        exit_szstr(&tmp_out);

        /* Variante 1 */

        tmp_in.data[0] = ' ';
        memcpy(&tmp_in.data[1], _src->static_bin_data, _src->len);

        tmp_in.len = _src->len + 1;

        result = _base64_encode(&tmp_in, &tmp_out, &arg->value.string);
        if (!result) goto exit;

        strip_base64_modifier_output(&tmp_out, 1, binary++);

        exit_szstr(&tmp_out);

        /* Variante 2 */

        tmp_in.data[0] = ' ';
        tmp_in.data[2] = ' ';
        memcpy(&tmp_in.data[2], _src->static_bin_data, _src->len);

        tmp_in.len = _src->len + 2;

        result = _base64_encode(&tmp_in, &tmp_out, &arg->value.string);
        if (!result) goto exit;

        strip_base64_modifier_output(&tmp_out, 2, binary++);

        exit_szstr(&tmp_out);

 exit:

        exit_szstr(&tmp_in);

    }

    if (!result)
    {
        for (i = 0; i < *dcount; i++)
            exit_szstr(&(*dest)[i]);

        free(*dest);

        *dcount = 0;
        *dest = NULL;

    }

 bad_arg:

    return result;

}
